/* 
 * Copyright (C) 2009 RDC Semiconductor Co.,Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * For technical support : 
 *     <jason.lin@rdc.com.tw>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86Resources.h"
#include "xf86RAC.h"
#include "xf86cmap.h"
#include "compiler.h"
#include "mibstore.h"
#include "vgaHW.h"
#include "mipointer.h"
#include "micmap.h"

#include "fb.h"
#include "regionstr.h"
#include "xf86xv.h"
#include <X11/extensions/Xv.h>
#include "vbe.h"

#include "xf86PciInfo.h"
#include "xf86Pci.h"

/* framebuffer offscreen manager */
#include "xf86fbman.h"

/* include xaa includes */
#include "xaa.h"
#include "xaarop.h"

/* H/W cursor support */
#include "xf86Cursor.h"

/* Driver specific headers */
#include "rdc.h"

#ifdef Accel_2D
/* ROP Translation Table */
int RDCXAACopyROP[16] =
{
    ROP_0,               /* GXclear */
    ROP_DSa,             /* GXand */
    ROP_SDna,            /* GXandReverse */
    ROP_S,               /* GXcopy */
    ROP_DSna,            /* GXandInverted */
    ROP_D,               /* GXnoop */
    ROP_DSx,             /* GXxor */
    ROP_DSo,             /* GXor */
    ROP_DSon,            /* GXnor */
    ROP_DSxn,            /* GXequiv */
    ROP_Dn,              /* GXinvert*/
    ROP_SDno,            /* GXorReverse */
    ROP_Sn,              /* GXcopyInverted */
    ROP_DSno,            /* GXorInverted */
    ROP_DSan,            /* GXnand */
    ROP_1                /* GXset */
};

int RDCXAAPatternROP[16]=
{
    ROP_0,
    ROP_DPa,
    ROP_PDna,
    ROP_P,
    ROP_DPna,
    ROP_D,
    ROP_DPx,
    ROP_DPo,
    ROP_DPon,
    ROP_PDxn,
    ROP_Dn,
    ROP_PDno,
    ROP_Pn,
    ROP_DPno,
    ROP_DPan,
    ROP_1
};

/* extern function */
extern void vWaitEngIdle(ScrnInfoPtr pScrn, RDCRecPtr pRDC);
extern UCHAR *pjRequestCMDQ(RDCRecPtr pRDC, ULONG ulDataLen);
extern Bool bGetLineTerm(_LINEInfo *LineInfo, LINEPARAM *dsLineParam);

/* Prototype type declaration */
Bool RDCAccelInit(ScreenPtr pScreen);
static void RDCSync(ScrnInfoPtr pScrn);
static void RDCSetupForScreenToScreenCopy(ScrnInfoPtr pScrn, 
                                          int xdir, int ydir, int rop,
                                          unsigned int planemask, int trans_color);
static void RDCSubsequentScreenToScreenCopy(ScrnInfoPtr pScrn, int x1, int y1, int x2,
                                            int y2, int w, int h);                                
static void RDCSetupForSolidFill(ScrnInfoPtr pScrn,
                                 int color, int rop, unsigned int planemask);
static void RDCSubsequentSolidFillRect(ScrnInfoPtr pScrn,
                                       int dst_x, int dst_y, int width, int height);  
static void RDCSetupForSolidLine(ScrnInfoPtr pScrn,
                                 int color, int rop, unsigned int planemask);
static void RDCSubsequentSolidHorVertLine(ScrnInfoPtr pScrn,
                                          int x, int y, int len, int dir);
static void RDCSetupForDashedLine(ScrnInfoPtr pScrn,
                                  int fg, int bg, int rop, unsigned int planemask,
                                  int length, UCHAR *pattern);
static void RDCSetupForMonoPatternFill(ScrnInfoPtr pScrn,
                                       int patx, int paty, int fg, int bg,
                                       int rop, unsigned int planemask);
static void RDCSubsequentMonoPatternFill(ScrnInfoPtr pScrn,
                                         int patx, int paty,
                                         int x, int y, int w, int h);
static void RDCSetupForColor8x8PatternFill(ScrnInfoPtr pScrn, int patx, int paty,
                               int rop, unsigned int planemask, int trans_col);
static void RDCSubsequentColor8x8PatternFillRect(ScrnInfoPtr pScrn, int patx, int paty,
                                                 int x, int y, int w, int h);
static void RDCSetupForCPUToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                                  int fg, int bg,
                                                  int rop, unsigned int planemask);
static void RDCSubsequentCPUToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                                    int x, int y,
                                                    int width, int height, int skipleft);
static void RDCSetupForScreenToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                                     int fg, int bg,
                                                     int rop, unsigned int planemask);
static void RDCSubsequentScreenToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                                       int x, int y, int width, int height,
                                                       int src_x, int src_y, int offset);
static void RDCSetClippingRectangle(ScrnInfoPtr pScrn,
                                    int left, int top, int right, int bottom);
static void RDCDisableClipping(ScrnInfoPtr pScrn); 

static void RDCSubsequentSolidTwoPointLine(ScrnInfoPtr pScrn,
                                           int x1, int y1, int x2, int y2, int flags);                                       
static void RDCSubsequentDashedTwoPointLine(ScrnInfoPtr pScrn,
                                            int x1, int y1, int x2, int y2,
                                            int flags, int phase);

Bool
RDCAccelInit(ScreenPtr pScreen)
{
    XAAInfoRecPtr  infoPtr;
    ScrnInfoPtr    pScrn = xf86Screens[pScreen->myNum];
    RDCRecPtr      pRDC = RDCPTR(pScrn);
    
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(0, X_INFO, DefaultLevel, "==Enter RDCAccelInit()== \n");
#endif
    pRDC->AccelInfoPtr = infoPtr = XAACreateInfoRec();
    if (!infoPtr)
    {
#if Accel_2D_DEBUG
        xf86DrvMsgVerb(0, X_INFO, ErrorLevel, "==Exit1 RDCAccelInit(), return FALSE== \n");
#endif
        return FALSE;
    }
    infoPtr->Flags = LINEAR_FRAMEBUFFER |
                     OFFSCREEN_PIXMAPS |
                     PIXMAP_CACHE;

    /* Sync */
    if (pRDC->ENGCaps & ENG_CAP_Sync)
        infoPtr->Sync = RDCSync;

    /* Screen To Screen copy */
    if (pRDC->ENGCaps & ENG_CAP_ScreenToScreenCopy)
    {    
        infoPtr->SetupForScreenToScreenCopy =  RDCSetupForScreenToScreenCopy;
        infoPtr->SubsequentScreenToScreenCopy = RDCSubsequentScreenToScreenCopy;
        infoPtr->ScreenToScreenCopyFlags = NO_TRANSPARENCY | NO_PLANEMASK;
    }

    /* Solid fill */
    if (pRDC->ENGCaps & ENG_CAP_SolidFill)
    {    
        infoPtr->SetupForSolidFill = RDCSetupForSolidFill;
        infoPtr->SubsequentSolidFillRect = RDCSubsequentSolidFillRect;
        infoPtr->SolidFillFlags = NO_PLANEMASK;
    } 
        
    /* Solid Lines */
    if (pRDC->ENGCaps & ENG_CAP_SolidLine)
    {    
        infoPtr->SetupForSolidLine = RDCSetupForSolidLine;            
        infoPtr->SubsequentSolidTwoPointLine = RDCSubsequentSolidTwoPointLine;
        infoPtr->SubsequentSolidHorVertLine = RDCSubsequentSolidHorVertLine;    
        infoPtr->SolidLineFlags = NO_PLANEMASK;
    }

    /* Dashed Lines */
    if (pRDC->ENGCaps & ENG_CAP_DashedLine)
    {
        infoPtr->SetupForDashedLine = RDCSetupForDashedLine;        
        infoPtr->SubsequentDashedTwoPointLine = RDCSubsequentDashedTwoPointLine;
        infoPtr->DashPatternMaxLength = 64;
        infoPtr->DashedLineFlags = NO_PLANEMASK |
                       LINE_PATTERN_MSBFIRST_LSBJUSTIFIED;
    }                                              

    /* 8x8 mono pattern fill */
    if (pRDC->ENGCaps & ENG_CAP_Mono8x8PatternFill)
    {    
        infoPtr->SetupForMono8x8PatternFill = RDCSetupForMonoPatternFill;
        infoPtr->SubsequentMono8x8PatternFillRect = RDCSubsequentMonoPatternFill;
        infoPtr->Mono8x8PatternFillFlags = NO_PLANEMASK |
                                           NO_TRANSPARENCY |
                                           HARDWARE_PATTERN_SCREEN_ORIGIN |
                                           HARDWARE_PATTERN_PROGRAMMED_BITS |
                                           BIT_ORDER_IN_BYTE_MSBFIRST;
    }                           
                           
    /* 8x8 color pattern fill */
    if (pRDC->ENGCaps & ENG_CAP_Color8x8PatternFill)
    {    
        infoPtr->SetupForColor8x8PatternFill = RDCSetupForColor8x8PatternFill;
        infoPtr->SubsequentColor8x8PatternFillRect = RDCSubsequentColor8x8PatternFillRect;
        infoPtr->Color8x8PatternFillFlags = NO_PLANEMASK |
                                            NO_TRANSPARENCY |    
                                            HARDWARE_PATTERN_SCREEN_ORIGIN;
    }                         
                     
    /* CPU To Screen Color Expand */
    if (pRDC->ENGCaps & ENG_CAP_CPUToScreenColorExpand)
    {    
        infoPtr->SetupForCPUToScreenColorExpandFill = RDCSetupForCPUToScreenColorExpandFill;
        infoPtr->SubsequentCPUToScreenColorExpandFill = RDCSubsequentCPUToScreenColorExpandFill;
        infoPtr->ColorExpandRange = MAX_PATReg_Size;
        infoPtr->ColorExpandBase = MMIOREG_PAT;
        infoPtr->CPUToScreenColorExpandFillFlags = NO_PLANEMASK |
                                                   BIT_ORDER_IN_BYTE_MSBFIRST;
    }                                  

    /* Screen To Screen Color Expand */
    if (pRDC->ENGCaps & ENG_CAP_ScreenToScreenColorExpand)
    {    
        infoPtr->SetupForScreenToScreenColorExpandFill = RDCSetupForScreenToScreenColorExpandFill;
        infoPtr->SubsequentScreenToScreenColorExpandFill = RDCSubsequentScreenToScreenColorExpandFill;
        infoPtr->ScreenToScreenColorExpandFillFlags = NO_PLANEMASK |
                                                      BIT_ORDER_IN_BYTE_MSBFIRST;
    }                                                  
                                              
    /* Clipping */
    if (pRDC->ENGCaps & ENG_CAP_Clipping)
    {    
        infoPtr->SetClippingRectangle = RDCSetClippingRectangle;
        infoPtr->DisableClipping = RDCDisableClipping;
        infoPtr->ClippingFlags = HARDWARE_CLIP_SCREEN_TO_SCREEN_COPY     |
                                 HARDWARE_CLIP_MONO_8x8_FILL |
                                 HARDWARE_CLIP_COLOR_8x8_FILL |
                                 HARDWARE_CLIP_SOLID_LINE | 
                                 HARDWARE_CLIP_DASHED_LINE | 
                                 HARDWARE_CLIP_SOLID_LINE; 
    }                         

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(0, X_INFO, DefaultLevel, "==Exit2 RDCAccelInit()== \n");
#endif

    return(XAAInit(pScreen, infoPtr));
} /* end of RDCAccelInit */


static void
RDCSync(ScrnInfoPtr pScrn)
{
    RDCRecPtr pRDC = RDCPTR(pScrn);

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(0, X_INFO, DefaultLevel, "==Enter RDCSync()== \n");
#endif
    /* wait engle idle */
    vWaitEngIdle(pScrn, pRDC);
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(0, X_INFO, DefaultLevel, "==Exit RDCSync()== \n");
#endif
} /* end of RDCSync */


static void RDCSetupForScreenToScreenCopy(ScrnInfoPtr pScrn,
                                          int xdir, int ydir, int rop,
                                          unsigned int planemask, int trans_color)
{
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG  cmdreg = 0;
    
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForScreenToScreenCopy(xdir = %d, ydir = %d, rop = 0x%x, planemask = 0x%x, trans_color = 0x%x)== \n", xdir, ydir, rop, planemask, trans_color);
#endif    
    /* Modify Reg. Value */
    cmdreg = CMD_BITBLT;
    switch (pRDC->VideoModeInfo.bitsPerPixel)
    {
        case 8:
            cmdreg |= CMD_COLOR_08;
            break;
        case 15:
        case 16:
            cmdreg |= CMD_COLOR_16;
            break;    
        case 24:
        case 32:
            cmdreg |= CMD_COLOR_32;
            break;        
    }
    
    cmdreg |= (RDCXAACopyROP[rop] << 8);
    
    if (xdir == -1)
    {
        cmdreg |= CMD_X_DEC;
    }
    if (ydir == -1)
    {
        cmdreg |= CMD_Y_DEC;
    }
    
    pRDC->ulCMDReg = cmdreg;
     
    if (!pRDC->MMIO2D)
    {   
        /* Write to CMDQ */
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*2);
                
        RDCSetupSRCPitch(pSingleCMD, pRDC->VideoModeInfo.ScreenPitch);  
        pSingleCMD++;
        RDCSetupDSTPitchHeight(pSingleCMD, pRDC->VideoModeInfo.ScreenPitch, -1);
    }
    else
    {
        /* Write to MMIO */                
        RDCSetupSRCPitch_MMIO(pRDC->VideoModeInfo.ScreenPitch);  
        RDCSetupDSTPitchHeight_MMIO(pRDC->VideoModeInfo.ScreenPitch, -1);    
    }

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSetupForScreenToScreenCopy()== \n");
#endif
} /* end of RDCSetupForScreenToScreenCopy */

static void
RDCSubsequentScreenToScreenCopy(ScrnInfoPtr pScrn, int x1, int y1, int x2,
                                int y2, int width, int height)
{
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    int src_x, src_y, dst_x, dst_y;
    ULONG srcbase, dstbase, cmdreg;

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSubsequentScreenToScreenCopy(x1 = %d, y1 = %d, x2 = %d, y2 = %d, width = %d, height = %d)== \n", x1, y1, x2, y2, width, height);
#endif        
    /* Modify Reg. Value */
    cmdreg = pRDC->ulCMDReg;

    if (pRDC->EnableClip)
        cmdreg |= CMD_ENABLE_CLIP;
    else
        cmdreg &= ~CMD_ENABLE_CLIP;        
        
    srcbase = pRDC->VideoModeInfo.ScreenPitch*y1;
    dstbase = pRDC->VideoModeInfo.ScreenPitch*y2;
          
    if (cmdreg & CMD_X_DEC)
    {
        src_x = x1 + width - 1;
        dst_x = x2 + width - 1;
    }
    else
    {
        src_x = x1;
        dst_x = x2;    
    }

    if (cmdreg & CMD_Y_DEC)
    {
        src_y = height - 1;
        dst_y = height - 1;
    }
    else
    {
        src_y = 0;
        dst_y = 0;    
    }
        
    if (!pRDC->MMIO2D)        
    {
        /* Write to CMDQ */
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*6);
      
        RDCSetupSRCBase(pSingleCMD, srcbase);
        pSingleCMD++;
        RDCSetupDSTBase(pSingleCMD, dstbase);
        pSingleCMD++;
        RDCSetupDSTXY(pSingleCMD, dst_x, dst_y);
        pSingleCMD++;
        RDCSetupSRCXY(pSingleCMD, src_x, src_y);
        pSingleCMD++;
        RDCSetupRECTXY(pSingleCMD, width, height);
        pSingleCMD++;
        RDCSetupCMDReg(pSingleCMD, cmdreg);

        /* Update Write Pointer */
        mUpdateWritePointer;
    
    }
    else
    {
        RDCSetupSRCBase_MMIO(srcbase);
        RDCSetupDSTBase_MMIO(dstbase);
        RDCSetupDSTXY_MMIO(dst_x, dst_y);    
        RDCSetupSRCXY_MMIO(src_x, src_y);    
        RDCSetupRECTXY_MMIO(width, height);    
        RDCSetupCMDReg_MMIO(cmdreg);       
                
        vWaitEngIdle(pScrn, pRDC);
    }
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSubsequentScreenToScreenCopy()== \n");
#endif
} /* end of RDCSubsequentScreenToScreenCopy */

static void
RDCSetupForSolidFill(ScrnInfoPtr pScrn,
                     int color, int rop, unsigned int planemask)
{
    
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG cmdreg;

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForSolidFill(color = 0x%x, rop = 0x%x, planemake = 0x%x)== \n", color, rop, planemask);
#endif
    /* Modify Reg. Value */
    cmdreg = CMD_BITBLT | CMD_PAT_FGCOLOR;
    switch (pRDC->VideoModeInfo.bitsPerPixel)
    {
    case 8:
        cmdreg |= CMD_COLOR_08;
        break;
    case 15:
    case 16:
        cmdreg |= CMD_COLOR_16;
        break;    
    case 24:
    case 32:
        cmdreg |= CMD_COLOR_32;
        break;        
    }
    cmdreg |= (RDCXAAPatternROP[rop] << 8);
    pRDC->ulCMDReg = cmdreg;
            
    if (!pRDC->MMIO2D)                    
    {
        /* Write to CMDQ */
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*2);

        RDCSetupDSTPitchHeight(pSingleCMD, pRDC->VideoModeInfo.ScreenPitch, -1);
        pSingleCMD++;
        RDCSetupFG(pSingleCMD, color);        
    }
    else
    {
        RDCSetupDSTPitchHeight_MMIO(pRDC->VideoModeInfo.ScreenPitch, -1);
        RDCSetupFG_MMIO(color);                
    }

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSetupForSolidFill()== \n");           
#endif
} /* end of RDCSetupForSolidFill */


static void
RDCSubsequentSolidFillRect(ScrnInfoPtr pScrn,
                           int dst_x, int dst_y, int width, int height)
{
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG dstbase, cmdreg;        

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSubsequentSolidFillRect(dst_x = %d, dst_y = %d, width = %d, height = %d)== \n", dst_x, dst_y, width, height);           
#endif
    /* Modify Reg. Value */
    cmdreg = pRDC->ulCMDReg;
    if (pRDC->EnableClip)
        cmdreg |= CMD_ENABLE_CLIP;
    else
        cmdreg &= ~CMD_ENABLE_CLIP;            
    dstbase = 0;

    if (dst_y >= pScrn->virtualY) 
    {   
        dstbase=pRDC->VideoModeInfo.ScreenPitch*dst_y;
        dst_y=0;
    }
                  
    if (!pRDC->MMIO2D)                    
    {                  
        /* Write to CMDQ */    
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*4);

        RDCSetupDSTBase(pSingleCMD, dstbase);
        pSingleCMD++;    
        RDCSetupDSTXY(pSingleCMD, dst_x, dst_y);
        pSingleCMD++;    
        RDCSetupRECTXY(pSingleCMD, width, height);
        pSingleCMD++;    
        RDCSetupCMDReg(pSingleCMD, cmdreg);        
      
        /* Update Write Pointer */
        mUpdateWritePointer;
                
    }
    else
    {                  
        RDCSetupDSTBase_MMIO(dstbase);
        RDCSetupDSTXY_MMIO(dst_x, dst_y);
        RDCSetupRECTXY_MMIO(width, height);
        RDCSetupCMDReg_MMIO(cmdreg);        
 
        vWaitEngIdle(pScrn, pRDC);
      
    }

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSubsequentSolidFillRect()== \n");           
#endif
} /* end of RDCSubsequentSolidFillRect */

/* Line */
static void RDCSetupForSolidLine(ScrnInfoPtr pScrn, 
                                 int color, int rop, unsigned int planemask)
{

    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG  cmdreg;
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==RDCSetupForSolidLine(color = 0x%x, rop = 0x%x, planemask = 0x%x)\n==", color, rop, planemask);
#endif
    /* Modify Reg. Value */
    cmdreg = CMD_BITBLT;
    
    switch (pRDC->VideoModeInfo.bitsPerPixel)
    {
    case 8:
        cmdreg |= CMD_COLOR_08;
        break;
    case 15:
    case 16:
        cmdreg |= CMD_COLOR_16;
        break;    
    case 24:
    case 32:
        cmdreg |= CMD_COLOR_32;
        break;        
    }
    cmdreg |= (RDCXAAPatternROP[rop] << 8);  
    
    pRDC->ulCMDReg = cmdreg;
     
    if (!pRDC->MMIO2D)
    {   
        /* Write to CMDQ */
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*3);

        RDCSetupDSTPitchHeight(pSingleCMD, pRDC->VideoModeInfo.ScreenPitch, -1);
        pSingleCMD++;
        RDCSetupFG(pSingleCMD, color);  
        pSingleCMD++;
        RDCSetupBG(pSingleCMD, 0);
        
    }
    else
    {
        /* Write to MMIO */   
        RDCSetupDSTPitchHeight_MMIO(pRDC->VideoModeInfo.ScreenPitch, -1);                     
        RDCSetupFG_MMIO(color);  
        RDCSetupBG_MMIO(0);            
    }
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSetupForSolidLine()== \n");           
#endif
} /* end of RDCSetupForSolidLine */


static void RDCSubsequentSolidHorVertLine(ScrnInfoPtr pScrn,
                                          int x, int y, int len, int dir)
{

    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG dstbase, cmdreg;   
    int width, height;
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSubsequentSolidHorVertLine(x = %d, y = %d, len = %d, dir = %d)\n==", x, y, len, dir);
#endif
            
    /* Modify Reg. Value */
    cmdreg = (pRDC->ulCMDReg & (~CMD_MASK)) | CMD_BITBLT;
    if (pRDC->EnableClip)
        cmdreg |= CMD_ENABLE_CLIP;
    else
        cmdreg &= ~CMD_ENABLE_CLIP;            
    dstbase = 0;
    
    if(dir == DEGREES_0)
    {            /* horizontal */
        width  = len;
        height = 1;    
    } 
    else 
    {                    /* vertical */
        width  = 1;
        height = len;            
    }
              
    if ((y + height) >= pScrn->virtualY) 
    {   
        dstbase=pRDC->VideoModeInfo.ScreenPitch*y;
        y=0;
    }
              
    
    if (!pRDC->MMIO2D)                    
    {                  
        /* Write to CMDQ */    
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*4);

        RDCSetupDSTBase(pSingleCMD, dstbase);
        pSingleCMD++;    
        RDCSetupDSTXY(pSingleCMD, x, y);
        pSingleCMD++;    
        RDCSetupRECTXY(pSingleCMD, width, height);
        pSingleCMD++;    
        RDCSetupCMDReg(pSingleCMD, cmdreg);        
      
        /* Update Write Pointer */
        mUpdateWritePointer;
               
    }
    else
    {                  
        RDCSetupDSTBase_MMIO(dstbase);
        RDCSetupDSTXY_MMIO(x, y);
        RDCSetupRECTXY_MMIO(width, height);
        RDCSetupCMDReg_MMIO(cmdreg);        
 
        vWaitEngIdle(pScrn, pRDC);
    }

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSubsequentSolidHorVertLine()== \n");           
#endif
} /* end of RDCSubsequentSolidHorVertLine */

/* Dash Line */
static void
RDCSetupForDashedLine(ScrnInfoPtr pScrn,
                      int fg, int bg, int rop, unsigned int planemask,
                      int length, UCHAR *pattern)
{

    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG  cmdreg;
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForDashedLine(fg = 0x%x, bg = 0x%x, rop = 0x%x, planemask = 0x%x, length = %d)\n==", fg, bg, rop, planemask, length);
#endif
    /* Modify Reg. Value */
    cmdreg = CMD_LINEDRAW | CMD_RESET_STYLE_COUNTER | CMD_ENABLE_LINE_STYLE;
    
    switch (pRDC->VideoModeInfo.bitsPerPixel)
    {
    case 8:
        cmdreg |= CMD_COLOR_08;
        break;
    case 15:
    case 16:
        cmdreg |= CMD_COLOR_16;
        break;    
    case 24:
    case 32:
        cmdreg |= CMD_COLOR_32;
        break;        
    }
    
    cmdreg |= (RDCXAAPatternROP[rop] << 8);  

    if(bg == -1) 
    {
        cmdreg |= CMD_TRANSPARENT;    
        bg = 0;
    }

    cmdreg |= (((length-1) & 0x3F) << 24);        /* line period */
    pRDC->ulCMDReg = cmdreg;
     
    if (!pRDC->MMIO2D)
    {   
        /* Write to CMDQ */
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*5);

        RDCSetupDSTPitchHeight(pSingleCMD, pRDC->VideoModeInfo.ScreenPitch, -1);
        pSingleCMD++;
        RDCSetupFG(pSingleCMD, fg);  
        pSingleCMD++;
        RDCSetupBG(pSingleCMD, bg); 
        pSingleCMD++;
        RDCSetupLineStyle1(pSingleCMD, *pattern);
        pSingleCMD++;
        RDCSetupLineStyle2(pSingleCMD, *(pattern+4));                   
                       
    }
    else
    {
        /* Write to MMIO */   
        RDCSetupDSTPitchHeight_MMIO(pRDC->VideoModeInfo.ScreenPitch, -1);
        RDCSetupFG_MMIO(fg);  
        RDCSetupBG_MMIO(bg); 
        RDCSetupLineStyle1_MMIO(*pattern);
        RDCSetupLineStyle2_MMIO(*(pattern+4));                
    }
                                     
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSetupForDashedLine()== \n");           
#endif
}

/* Mono Pattern Fill */
static void
RDCSetupForMonoPatternFill(ScrnInfoPtr pScrn,
                           int patx, int paty, int fg, int bg,
                           int rop, unsigned int planemask)
{
    
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG cmdreg;

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFill(patx = %d, paty = %d, fg = 0x%x, bg = 0x%x, rop = 0x%x, planemask = 0x%x)\n==", patx, paty, fg, bg, rop, planemask);
#endif
    /* Modify Reg. Value */
    cmdreg = CMD_BITBLT | CMD_PAT_MONOMASK;
    switch (pRDC->VideoModeInfo.bitsPerPixel)
    {
    case 8:
        cmdreg |= CMD_COLOR_08;
        break;
    case 15:
    case 16:
        cmdreg |= CMD_COLOR_16;
        break;    
    case 24:
    case 32:
        cmdreg |= CMD_COLOR_32;
        break;        
    }
    
    cmdreg |= (RDCXAAPatternROP[rop] << 8);
    pRDC->ulCMDReg = cmdreg;
            
    if (!pRDC->MMIO2D)                    
    {
        /* Write to CMDQ */
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*5);

        RDCSetupDSTPitchHeight(pSingleCMD, pRDC->VideoModeInfo.ScreenPitch, -1);
        pSingleCMD++;
        RDCSetupFG(pSingleCMD, fg);  
        pSingleCMD++;
        RDCSetupBG(pSingleCMD, bg);
        pSingleCMD++;
        RDCSetupMONO1(pSingleCMD, patx);  
        pSingleCMD++;
        RDCSetupMONO2(pSingleCMD, paty);                           
    }
    else
    {
        RDCSetupDSTPitchHeight_MMIO(pRDC->VideoModeInfo.ScreenPitch, -1);
        RDCSetupFG_MMIO(fg);   
        RDCSetupBG_MMIO(bg);
        RDCSetupMONO1_MMIO(patx);  
        RDCSetupMONO2_MMIO(paty);                         
    }
               
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSetupForMonoPatternFill()== \n");           
#endif
} /* end of RDCSetupForMonoPatternFill */

                      
static void
RDCSubsequentMonoPatternFill(ScrnInfoPtr pScrn,
                             int patx, int paty,
                             int dst_x, int dst_y, int width, int height)
{
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG dstbase, cmdreg;        
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCSubsequentMonoPatternFill(patx = %d, paty = %d, dst_x = %d, dst_y = %d, width = %d, height = %d,)\n==", patx, paty, dst_x, dst_y, width, height);
#endif
            
    /* Modify Reg. Value */
    cmdreg = pRDC->ulCMDReg;
    if (pRDC->EnableClip)
        cmdreg |= CMD_ENABLE_CLIP;
    else
        cmdreg &= ~CMD_ENABLE_CLIP;            
    dstbase = 0;

    if (dst_y >= pScrn->virtualY) 
    {   
        dstbase=pRDC->VideoModeInfo.ScreenPitch*dst_y;
        dst_y=0;
    }
                  
    if (!pRDC->MMIO2D)                    
    {                  
        /* Write to CMDQ */    
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*4);

        RDCSetupDSTBase(pSingleCMD, dstbase);
        pSingleCMD++;    
        RDCSetupDSTXY(pSingleCMD, dst_x, dst_y);
        pSingleCMD++;    
        RDCSetupRECTXY(pSingleCMD, width, height);
        pSingleCMD++;    
        RDCSetupCMDReg(pSingleCMD, cmdreg);        
      
        /* Update Write Pointer */
        mUpdateWritePointer;
              
    }
    else
    {                  
        RDCSetupDSTBase_MMIO(dstbase);
        RDCSetupDSTXY_MMIO(dst_x, dst_y);
        RDCSetupRECTXY_MMIO(width, height);
        RDCSetupCMDReg_MMIO(cmdreg);        
 
        vWaitEngIdle(pScrn, pRDC);      
    }    

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSubsequentMonoPatternFill()== \n");           
#endif
} /* end of RDCSubsequentMonoPatternFill */

static void
RDCSetupForColor8x8PatternFill(ScrnInfoPtr pScrn, int patx, int paty,
                   int rop, unsigned int planemask, int trans_col)
{
    
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG cmdreg;
    CARD32 *pataddr;
    ULONG ulPatSize;
    int i, j, cpp;
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCSetupForColor8x8PatternFill(patx = %d, paty = %d, rop = 0x%x, planemask = 0x%x, trans_col = 0x%x)\n==", patx, paty, rop, planemask, trans_col);
#endif
    /* Modify Reg. Value */
    cmdreg = CMD_BITBLT | CMD_PAT_PATREG;

    switch (pRDC->VideoModeInfo.bitsPerPixel)
    {
    case 8:
        cmdreg |= CMD_COLOR_08;
        break;
    case 15:
    case 16:
        cmdreg |= CMD_COLOR_16;
        break;    
    case 24:
    case 32:
        cmdreg |= CMD_COLOR_32;
        break;        
    }
    
    cmdreg |= (RDCXAAPatternROP[rop] << 8);
    pRDC->ulCMDReg = cmdreg;
    cpp = (pScrn->bitsPerPixel + 1) / 8;
    pataddr = (CARD32 *)(pRDC->FBVirtualAddr +
                        (paty * pRDC->VideoModeInfo.ScreenWidth) + (patx * cpp));   
    ulPatSize = 8*8*cpp;
                
    if (!pRDC->MMIO2D)                    
    {
        /* Write to CMDQ */        
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*(1 + ulPatSize/4));
        RDCSetupDSTPitchHeight(pSingleCMD, pRDC->VideoModeInfo.ScreenPitch, -1);
        pSingleCMD++;
        for (i=0; i<8; i++)
        {
            for (j=0; j<8*cpp/4; j++)
            {
                RDCSetupPatReg(pSingleCMD, (i*j + j) , (*(CARD32 *) (pataddr++)));
                pSingleCMD++;                    
            }    
        }                
    }
    else
    {        
        RDCSetupDSTPitchHeight_MMIO(pRDC->VideoModeInfo.ScreenPitch, -1);
        for (i=0; i<8; i++)
        {
            for (j=0; j<8*cpp/4; j++)
            {
                RDCSetupPatReg_MMIO((i*j + j) , (*(CARD32 *) (pataddr++)));
            }    
        }                
                 
    }

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSetupForColor8x8PatternFill()== \n");           
#endif
} /* end of RDCSetupForColor8x8PatternFill */
               
static void
RDCSubsequentColor8x8PatternFillRect(ScrnInfoPtr pScrn, int patx, int paty,
                                     int dst_x, int dst_y, int width, int height)
{
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG dstbase, cmdreg;        
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCSubsequentColor8x8PatternFillRect(patx = %d, paty = %d, dst_x = %d, dst_y = %d, width = %d, height)\n==", patx, paty, dst_x, dst_y, width, height);
#endif
            
    /* Modify Reg. Value */
    cmdreg = pRDC->ulCMDReg;
    if (pRDC->EnableClip)
        cmdreg |= CMD_ENABLE_CLIP;
    else
        cmdreg &= ~CMD_ENABLE_CLIP;            

    dstbase = 0;

    if (dst_y >= pScrn->virtualY) 
    {   
        dstbase=pRDC->VideoModeInfo.ScreenPitch*dst_y;
        dst_y=0;
    }
                  
    if (!pRDC->MMIO2D)                    
    {                  
        /* Write to CMDQ */    
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*4);

        RDCSetupDSTBase(pSingleCMD, dstbase);
        pSingleCMD++;    
        RDCSetupDSTXY(pSingleCMD, dst_x, dst_y);
        pSingleCMD++;    
        RDCSetupRECTXY(pSingleCMD, width, height);
        pSingleCMD++;    
        RDCSetupCMDReg(pSingleCMD, cmdreg);        
      
        /* Update Write Pointer */
        mUpdateWritePointer;
                
    }
    else
    {                  
        RDCSetupDSTBase_MMIO(dstbase);
        RDCSetupDSTXY_MMIO(dst_x, dst_y);
        RDCSetupRECTXY_MMIO(width, height);
        RDCSetupCMDReg_MMIO(cmdreg);        
 
        vWaitEngIdle(pScrn, pRDC);      
    }    
    
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSubsequentColor8x8PatternFillRect()== \n");           
#endif
} /* RDCSubsequentColor8x8PatternFillRect */

/* CPU to Screen Expand */
static void
RDCSetupForCPUToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                      int fg, int bg,
                                      int rop, unsigned int planemask)
{

    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG cmdreg;

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCSetupForCPUToScreenColorExpandFill(fg = 0x%x, bg = ox%x, rop = ox%x, planemask = ox%x)\n==", fg, bg, rop, planemask);
#endif
    /* Modify Reg. Value */
    cmdreg = CMD_COLOREXP;

    switch (pRDC->VideoModeInfo.bitsPerPixel)
    {
    case 8:
        cmdreg |= CMD_COLOR_08;
        break;
    case 15:
    case 16:
        cmdreg |= CMD_COLOR_16;
        break;    
    case 24:
    case 32:
        cmdreg |= CMD_COLOR_32;
        break;        
    }
    
    cmdreg |= (RDCXAAPatternROP[rop] << 8);

    if(bg == -1) 
    {
        cmdreg |= CMD_FONT_TRANSPARENT;    
        bg = 0;
    }
    
    pRDC->ulCMDReg = cmdreg;
            
    if (!pRDC->MMIO2D)                    
    {
        /* Write to CMDQ */
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*3);

        RDCSetupDSTPitchHeight(pSingleCMD, pRDC->VideoModeInfo.ScreenPitch, -1);
        pSingleCMD++;
        RDCSetupFG(pSingleCMD, fg);  
        pSingleCMD++;
        RDCSetupBG(pSingleCMD, bg); 
       
    }
    else
    {
        RDCSetupDSTPitchHeight_MMIO(pRDC->VideoModeInfo.ScreenPitch, -1);
        RDCSetupFG_MMIO(fg); 
        RDCSetupBG_MMIO(bg); 
                       
    }
               
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSetupForCPUToScreenColorExpandFill()== \n");           
#endif
}
                               
static void
RDCSubsequentCPUToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                        int dst_x, int dst_y,
                                        int width, int height, int offset)
{

    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG dstbase, cmdreg;

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCSetupForCPUToScreenColorExpandFill(dst_x = %d, dst_y = %d, width = %d, height, offset = %d)\n==", dst_x, dst_y, width, height, offset);
#endif
            
    /* Modify Reg. Value */
    cmdreg = pRDC->ulCMDReg;
    if (pRDC->EnableClip)
        cmdreg |= CMD_ENABLE_CLIP;
    else
        cmdreg &= ~CMD_ENABLE_CLIP;            
    dstbase = 0;

    if (dst_y >= pScrn->virtualY) 
    {   
        dstbase=pRDC->VideoModeInfo.ScreenPitch*dst_y;
        dst_y=0;
    }
                  
    if (!pRDC->MMIO2D)                    
    {                  
        /* Write to CMDQ */    
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*5);

        RDCSetupSRCPitch(pSingleCMD, ((width+7)/8));  
        pSingleCMD++;    
        RDCSetupDSTBase(pSingleCMD, dstbase);
        pSingleCMD++;    
        RDCSetupDSTXY(pSingleCMD, dst_x, dst_y);
        pSingleCMD++;    
        RDCSetupRECTXY(pSingleCMD, width, height);
        pSingleCMD++;    
        RDCSetupCMDReg(pSingleCMD, cmdreg);        
      
        /* Update Write Pointer */
        mUpdateWritePointer;
               
    }
    else
    {  
        RDCSetupSRCPitch_MMIO((width+7)/8);                          
        RDCSetupDSTBase_MMIO(dstbase);
        RDCSetupDSTXY_MMIO(dst_x, dst_y);
        RDCSetupSRCXY_MMIO(0, 0);
        
        RDCSetupRECTXY_MMIO(width, height);
        RDCSetupCMDReg_MMIO(cmdreg);        
 
        vWaitEngIdle(pScrn, pRDC);      
    }
        
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSubsequentCPUToScreenColorExpandFill()== \n");           
#endif
}


/* Screen to Screen Color Expand */
static void
RDCSetupForScreenToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                         int fg, int bg,
                                         int rop, unsigned int planemask)
{
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG cmdreg;

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCSetupForScreenToScreenColorExpandFill(fg = 0x%x, bg = ox%x, rop = ox%x, planemask = ox%x)\n==", fg, bg, rop, planemask);
#endif
           
    /* Modify Reg. Value */
    cmdreg = CMD_ENHCOLOREXP;
    switch (pRDC->VideoModeInfo.bitsPerPixel)
    {
    case 8:
        cmdreg |= CMD_COLOR_08;
        break;
    case 15:
    case 16:
        cmdreg |= CMD_COLOR_16;
        break;    
    case 24:
    case 32:
        cmdreg |= CMD_COLOR_32;
        break;        
    }
    
    cmdreg |= (RDCXAAPatternROP[rop] << 8);
    
    if(bg == -1)
    {
        cmdreg |= CMD_FONT_TRANSPARENT;    
        bg = 0;
    }    
    pRDC->ulCMDReg = cmdreg;
            
    if (!pRDC->MMIO2D)                    
    {
        /* Write to CMDQ */
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*3);

        RDCSetupDSTPitchHeight(pSingleCMD, pRDC->VideoModeInfo.ScreenPitch, -1);
        pSingleCMD++;
        RDCSetupFG(pSingleCMD, fg);  
        pSingleCMD++;
        RDCSetupBG(pSingleCMD, bg); 
       
    }
    else
    {
        RDCSetupDSTPitchHeight_MMIO(pRDC->VideoModeInfo.ScreenPitch, -1);
        RDCSetupFG_MMIO(fg); 
        RDCSetupBG_MMIO(bg); 
                       
    }
               
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSetupForScreenToScreenColorExpandFill()== \n");           
#endif
}



static void
RDCSubsequentScreenToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                           int dst_x, int dst_y, int width, int height,
                                           int src_x, int src_y, int offset)
{
   RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
    ULONG srcbase, dstbase, cmdreg;
    USHORT srcpitch;

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCSubsequentScreenToScreenColorExpandFill(dst_x = %d, dst_y = %d, width = %d, height, src_x = %d, src_y = %d, offset = %d)\n==", dst_x, dst_y, width, height, src_x, src_y, offset);
#endif
            
    /* Modify Reg. Value */
    cmdreg = pRDC->ulCMDReg;
    
    if (pRDC->EnableClip)
        cmdreg |= CMD_ENABLE_CLIP;
    else
        cmdreg &= ~CMD_ENABLE_CLIP;   
        
    dstbase = 0;
    
    if (dst_y >= pScrn->virtualY) 
    {   
        dstbase=pRDC->VideoModeInfo.ScreenPitch*dst_y;
        dst_y=0;
    }
    
    srcbase = pRDC->VideoModeInfo.ScreenPitch*src_y + ((pScrn->bitsPerPixel+1)/8)*src_x;            
    srcpitch = (pScrn->displayWidth+7)/8;
    
    if (!pRDC->MMIO2D)                    
    {                  
        /* Write to CMDQ */    
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*6);

        RDCSetupSRCBase(pSingleCMD, srcbase);
        pSingleCMD++;  
        RDCSetupSRCPitch(pSingleCMD,srcpitch);  
        pSingleCMD++;    
        RDCSetupDSTBase(pSingleCMD, dstbase);
        pSingleCMD++;    
        RDCSetupDSTXY(pSingleCMD, dst_x, dst_y);
        pSingleCMD++;    
        RDCSetupRECTXY(pSingleCMD, width, height);
        pSingleCMD++;    
        RDCSetupCMDReg(pSingleCMD, cmdreg);        
      
        /* Update Write Pointer */
        mUpdateWritePointer;
       
    }
    else
    { 
        RDCSetupSRCBase_MMIO(srcbase);         
        RDCSetupSRCPitch_MMIO(srcpitch);                          
        RDCSetupDSTBase_MMIO(dstbase);
        RDCSetupDSTXY_MMIO(dst_x, dst_y);
        RDCSetupRECTXY_MMIO(width, height);
        RDCSetupCMDReg_MMIO(cmdreg);        
 
        vWaitEngIdle(pScrn, pRDC);
      
    }
        
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSubsequentScreenToScreenColorExpandFill()== \n");           
#endif
}

      
/* Clipping */
static void
RDCSetClippingRectangle(ScrnInfoPtr pScrn,
                        int left, int top, int right, int bottom)
{
    
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC *pSingleCMD;
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCSetClippingRectangle(left = %d, top = %d, right = %d, bottom = %d)\n==", left, top, right, bottom);
#endif
    pRDC->EnableClip = TRUE;
            
    if (!pRDC->MMIO2D)                    
    {
        /* Write to CMDQ */
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*2);

        RDCSetupCLIP1(pSingleCMD, left, top);
        pSingleCMD++;
        RDCSetupCLIP2(pSingleCMD, right, bottom);                 
    }
    else
    {
        RDCSetupCLIP1_MMIO(left, top);
        RDCSetupCLIP2_MMIO(right, bottom);                       
    }
    
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSetClippingRectangle()== \n");           
#endif
}

static void
RDCDisableClipping(ScrnInfoPtr pScrn)
{
    RDCRecPtr pRDC = RDCPTR(pScrn);
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCDisableClipping()\n==");
#endif
    pRDC->EnableClip = FALSE;
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCDisableClipping()== \n");           
#endif
}

static void RDCSubsequentSolidTwoPointLine(ScrnInfoPtr pScrn,
                                           int x1, int y1, int x2, int y2, int flags)
{
 
    RDCRecPtr pRDC = RDCPTR(pScrn);
    PKT_SC    *pSingleCMD;
    ULONG     dstbase, ulCommand;
    ULONG     miny, maxy;         
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCSubsequentSolidTwoPointLine(x1 = %d, y1 = %d, x2 = %d, y2 = %d, flags = 0x%x)\n==", x1, y1, x2, y2, flags);
#endif

    /* Modify Reg. Value */
    ulCommand = (pRDC->ulCMDReg & (~CMD_MASK)) | CMD_LINEDRAW | CMD_NORMAL_LINE;
    
    if(flags & OMIT_LAST)
        ulCommand |= CMD_NOT_DRAW_LAST_PIXEL;
    else
        ulCommand &= ~CMD_NOT_DRAW_LAST_PIXEL;
        
    if (pRDC->EnableClip)
        ulCommand |= CMD_ENABLE_CLIP;
    else
        ulCommand &= ~CMD_ENABLE_CLIP;        
    dstbase = 0;
    miny = (y1 > y2) ? y2 : y1;
    maxy = (y1 > y2) ? y1 : y2;
    
    if(maxy >= pScrn->virtualY) 
    {
        dstbase = pRDC->VideoModeInfo.ScreenPitch * miny;
        y1 -= miny;
        y2 -= miny;
    }

    if (!pRDC->MMIO2D)                    
    {                  
        /* Write to CMDQ */    
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*5);

        RDCSetupDSTBase(pSingleCMD, dstbase);
        pSingleCMD++;    
        RDCSetupLineXY(pSingleCMD, x1, y1);
        pSingleCMD++;
        RDCSetupLineXY2(pSingleCMD, x2, y2);
        pSingleCMD++;
        RDCSetupLineNumber(pSingleCMD, 0);
        pSingleCMD++;                     
        RDCSetupCMDReg(pSingleCMD, ulCommand);        
              
        /* Update Write Pointer */
        mUpdateWritePointer;                

        /* Patch KDE pass abnormal point, ycchen@052507 */
        vWaitEngIdle(pScrn, pRDC);
        
    }
    else
    {                  
        RDCSetupDSTBase_MMIO(dstbase);
        RDCSetupLineXY_MMIO(x1, y1);
        RDCSetupLineXY2_MMIO(x2, y2);
        RDCSetupLineNumber_MMIO(0);
        RDCSetupCMDReg_MMIO(ulCommand);        
             
        vWaitEngIdle(pScrn, pRDC);
    }

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSubsequentSolidTwoPointLine()== \n");           
#endif
} /* end of M2010SubsequentSolidTwoPointLine */

/* M2010 Dash Line */
static void
RDCSubsequentDashedTwoPointLine(ScrnInfoPtr pScrn,
                                int x1, int y1, int x2, int y2,
                                int flags, int phase)
{
 
    RDCRecPtr  pRDC = RDCPTR(pScrn);
    PKT_SC     *pSingleCMD;
    ULONG      dstbase, ulCommand; 
    ULONG      miny, maxy;  
#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Enter RDCSetupForMonoPatternFillRDCSubsequentDashedTwoPointLine(x1 = %d, y1 = %d, x2 = %d, y2 = %d, flags = 0x%x, phase = %d(0x%x)\n==", x1, y1, x2, y2, phase, phase);
#endif

    /* Modify Reg. Value */
    ulCommand = pRDC->ulCMDReg | CMD_NORMAL_LINE;
    
    if(flags & OMIT_LAST)
        ulCommand |= CMD_NOT_DRAW_LAST_PIXEL;
    else
        ulCommand &= ~CMD_NOT_DRAW_LAST_PIXEL;
        
    if (pRDC->EnableClip)
        ulCommand |= CMD_ENABLE_CLIP;
    else
        ulCommand &= ~CMD_ENABLE_CLIP;        
    dstbase = 0;        
    miny = (y1 > y2) ? y2 : y1;
    maxy = (y1 > y2) ? y1 : y2;
    
    if(maxy >= pScrn->virtualY) 
    {
        dstbase = pRDC->VideoModeInfo.ScreenPitch * miny;
        y1 -= miny;
        y2 -= miny;
    }

    if (!pRDC->MMIO2D)                    
    {                  
        /* Write to CMDQ */    
        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pRDC, PKT_SINGLE_LENGTH*5);

        RDCSetupDSTBase(pSingleCMD, dstbase);
        pSingleCMD++;    
        RDCSetupLineXY(pSingleCMD, x1, y1);
        pSingleCMD++;
        RDCSetupLineXY2(pSingleCMD, x2, y2);
        pSingleCMD++;
        RDCSetupLineNumber(pSingleCMD, 0);
        pSingleCMD++;                     
        RDCSetupCMDReg(pSingleCMD, ulCommand);        
              
        /* Update Write Pointer */
        mUpdateWritePointer;

        /* Patch KDE pass abnormal point, ycchen@052507 */
        vWaitEngIdle(pScrn, pRDC);
              
    }
    else
    {                  
        RDCSetupDSTBase_MMIO(dstbase);
        RDCSetupLineXY_MMIO(x1, y1);
        RDCSetupLineXY2_MMIO(x2, y2);
        RDCSetupLineNumber_MMIO(0);
        RDCSetupCMDReg_MMIO(ulCommand);
       
        vWaitEngIdle(pScrn, pRDC);
    }

#if Accel_2D_DEBUG
    xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DefaultLevel, "==Exit RDCSubsequentDashedTwoPointLine()== \n");           
#endif
}
                                         
#endif    /* end of Accel_2D */
